/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambuilderskin.h"
#include <gdam/component/gdambuilder.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkvseparator.h>
#include <gtk/gtktable.h>
#include "gdamskinutil.h"
#include "stdlib.h"
#include "string.h"
#include <glade/glade.h>

static GtkObjectClass* parent_class = NULL;

static void gdam_builder_skin_class_init(GdamBuilderSkinClass* klass);

static void gdam_builder_skin_synchronize(GdamView* comp_class);

static gboolean gdam_builder_skin_build_new_model(GdamSkin*    skin,
						      GdamChannel* channel,
						      int          num_options,
	                                              char**       options);
/*    ____ _                 ____  _          __  __ 
 *   / ___| | __ _ ___ ___  / ___|| |_ _   _ / _|/ _|
 *  | |   | |/ _` / __/ __| \___ \| __| | | | |_| |_ 
 *  | |___| | (_| \__ \__ \  ___) | |_| |_| |  _|  _|
 *   \____|_|\__,_|___/___/ |____/ \__|\__,_|_| |_|  
 */                                                 


static void gdam_builder_skin_init(GdamBuilderSkin* skin)
{
	skin->filter_boxes = g_hash_table_new(NULL, NULL);
}

GtkType gdam_builder_skin_get_type()
{
	static GtkType builderskin_type = 0;
	if (!builderskin_type) {
		static const GtkTypeInfo builderskin_info = {
			"GdamBuilderSkin",
			sizeof(GdamBuilderSkin),
			sizeof(GdamBuilderSkinClass),
			(GtkClassInitFunc) gdam_builder_skin_class_init,
			(GtkObjectInitFunc) gdam_builder_skin_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_SKIN;
		parent_class = gtk_type_class(parent);
		builderskin_type = gtk_type_unique(parent, &builderskin_info);
	}
	return builderskin_type;
}

typedef struct _FilterInfo {
	    GdamBuilderSkin* skin;
	    GdamFilter* filter;
	    GdamSkin* filter_skin;
	    GtkWidget* widget;
	    char* component_name;
	    char* skin_name;
	    int   num_options;
	    char** options;
} FilterInfo;

static void filter_destroy_callback(GtkWidget* widget, FilterInfo* fi)
{
	GdamSkin* skin = GDAM_SKIN(fi->skin);
	if (fi->filter)
		gdam_builder_remove_filter(GDAM_BUILDER(skin->model),
		                           fi->filter);
	gtk_widget_destroy(fi->widget);
	g_hash_table_remove(fi->skin->filter_boxes, fi->filter_skin);
	gtk_container_queue_resize(GTK_CONTAINER(
		gtk_widget_get_toplevel(GTK_WIDGET(fi->skin->filterbox))));
}
static void filter_toggle_disable_callback(GtkWidget* widget, FilterInfo* fi)
{
	GtkToggleButton* toggle_button = GTK_TOGGLE_BUTTON(widget);
	GdamBuilder* builder = GDAM_BUILDER(GDAM_SKIN(fi->skin)->model);
	if (toggle_button->active)
		gdam_builder_disable_filter(builder, fi->filter);
	else
		gdam_builder_enable_filter(builder, fi->filter);
}
static void filter_move_up_callback(GtkWidget* widget, FilterInfo* fi)
{
	gdam_builder_skin_move_up(fi->skin, fi->filter_skin);
}

static void filter_move_down_callback(GtkWidget* widget, FilterInfo* fi)
{
	gdam_builder_skin_move_down(fi->skin, fi->filter_skin);
}

static GtkWidget* make_widget_for_filter(GdamBuilderSkin* builder_skin,
                                         GdamSkin*        filter_skin,
				         FilterInfo*      filter_info)
{
	GtkWidget *hbox, *button, *disable_toggle;
	GtkWidget *move_up, *move_down, *bbox;
	GtkWidget* divider;

	if (filter_info == NULL) {
		filter_info = g_new0(FilterInfo, 1);
		filter_info->skin = builder_skin;
		filter_info->filter_skin = filter_skin;
		filter_info->filter = GDAM_FILTER(filter_skin->model);
	}
	filter_info->widget = hbox = gtk_hbox_new(FALSE,0);
	bbox = gtk_table_new(3, 2, FALSE);
	gtk_widget_show(bbox);
	divider = gtk_vseparator_new();
	gtk_widget_show(divider);
	gtk_table_attach_defaults(GTK_TABLE(bbox), divider, 2, 3, 0, 2);
	gtk_box_pack_start(GTK_BOX(hbox),
	                   bbox,
			   FALSE,
			   FALSE,
			   0);
	button = gtk_button_new_with_label("X");
	gtk_signal_connect(GTK_OBJECT(button),
			   "clicked",
			   GTK_SIGNAL_FUNC(filter_destroy_callback),
			   filter_info);
	gtk_table_attach_defaults(GTK_TABLE(bbox), button, 0, 1, 0, 1);
	disable_toggle = gtk_toggle_button_new_with_label("0");
	gtk_signal_connect(GTK_OBJECT(disable_toggle),
			   "clicked",
			   GTK_SIGNAL_FUNC(filter_toggle_disable_callback),
			   filter_info);
	gtk_table_attach_defaults(GTK_TABLE(bbox), disable_toggle, 1, 2, 0, 1);
	move_up = gtk_button_new_with_label("<");
	gtk_signal_connect(GTK_OBJECT(move_up),
			   "clicked",
			   GTK_SIGNAL_FUNC(filter_move_up_callback),
			   filter_info);
	gtk_table_attach_defaults(GTK_TABLE(bbox), move_up, 0, 1, 1, 2);

	move_down = gtk_button_new_with_label(">");
	gtk_signal_connect(GTK_OBJECT(move_down),
			   "clicked",
			   GTK_SIGNAL_FUNC(filter_move_down_callback),
			   filter_info);
	gtk_table_attach_defaults(GTK_TABLE(bbox), move_down, 1, 2, 1, 2);

	gtk_widget_ref(filter_skin->main_widget);
	if (filter_skin->main_widget->parent)
		gtk_container_remove(
			GTK_CONTAINER(filter_skin->main_widget->parent),
			filter_skin->main_widget
		);
	gtk_box_pack_start(GTK_BOX(hbox), filter_skin->main_widget, 
			   TRUE, TRUE, 0);
	gtk_widget_unref(filter_skin->main_widget);

	gtk_widget_show(button);
	gtk_widget_show(disable_toggle);
	gtk_widget_show(move_up);
	gtk_widget_show(move_down);
	gtk_widget_show(hbox);
	g_hash_table_insert(builder_skin->filter_boxes, 
	                    filter_skin,
			    hbox);
	return hbox;
}

static void filter_create_callback(GtkWidget* widget, FilterInfo* fi)
{
	GdamModel* filter;
	GdamSkin* filter_skin;
	GtkWidget* fwidget;
	GdamSkin* skin = GDAM_SKIN(fi->skin);
	GdamBuilder* builder = GDAM_BUILDER(skin->model);
	FilterInfo* filter_info;

	if (builder->component == NULL) {
		g_warning("builder: filter_create_callback: "
				"builder needs child to add filters to");
		return;
	}

	filter_skin = gdam_skin_new_by_type_name(fi->component_name, 
				                 fi->skin_name,
					         fi->num_options,
					         fi->options,
				            GDAM_MODEL(builder)->channel,
				                 &filter);
	if (filter_skin == NULL) {
		g_warning("error building %s/%s", fi->component_name, 
						  fi->skin_name);
		return;
	}
	if (!GDAM_IS_FILTER(filter)) {
		g_warning("sorry, %s is not a filter.", fi->component_name);
		return;
	}
	gdam_builder_append_filter(builder, GDAM_FILTER(filter), TRUE);

	filter_info = g_new(FilterInfo,1);
	memcpy(filter_info, fi, sizeof(FilterInfo));
	filter_info->filter = (GdamFilter*)filter;
	filter_info->filter_skin = filter_skin;
	filter_info->widget = make_widget_for_filter(filter_info->skin,
	                                             filter_skin,
					             filter_info);

	/* Add filter's main_widget to the filter_box, libglade
	 * may decide to parent the widget, so fix that... */
	fwidget = filter_info->widget;
	gtk_widget_ref(fwidget);
	if (fwidget->parent)
		gtk_container_remove(GTK_CONTAINER(fwidget->parent),
		                     fwidget);
	gtk_box_pack_start(fi->skin->filterbox,
			   fwidget,
			   FALSE, FALSE, 2);
	gtk_widget_unref(fwidget);

}
static gboolean gdam_builder_filter_box_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GDAM_BUILDER_SKIN(skin)->filterbox = GTK_BOX(object);
	return TRUE;
}

static gboolean gdam_builder_launch_filter_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             subs)
{
	FilterInfo* fi = g_new0(FilterInfo,1);
	if (num_handler_names != 3) {
		g_warning("too few arguments to %s", subs[0]);
		return TRUE;
	}
	fi->skin = GDAM_BUILDER_SKIN(skin);
	fi->component_name = subs[1];
	fi->skin_name = subs[2];
	subs[1] = subs[2] = NULL;
	gtk_signal_connect(
		object,
		signal_name,
		GTK_SIGNAL_FUNC(filter_create_callback),
		fi);
	return TRUE;
} 
static gboolean gdam_builder_set_child_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamBuilderSkin* builder_skin = GDAM_BUILDER_SKIN(skin);
	builder_skin->child_container = GTK_CONTAINER(object);
	if (builder_skin->child_skin) {
		gdam_debug("gdam_builder_set_child_handler");
		gdam_set_container(builder_skin->child_container,
				   builder_skin->child_skin->main_widget);
	}
	return TRUE;
}

gboolean gdam_builder_skin_set_child(GdamBuilderSkin* 	skin,
				     const char* 	skin_type_name, 
				     const char* 	skin_filename,
				     int                num_options,
				     char**             options)
{
	GdamModel* subcomponent;
	GdamSkin* subcomponent_skin;
	GdamBuilder* builder = GDAM_BUILDER(GDAM_SKIN(skin)->model);
	gdam_debug("gdam_builder_skin_set_child: "
	           "type=%s, filename=%s", 
		   skin_type_name, skin_filename);
	subcomponent_skin = gdam_skin_new_by_type_name
	                             (skin_type_name,
                                      skin_filename,
				      num_options,
				      options,
				      GDAM_MODEL(builder)->channel,
				      &subcomponent);
	if (subcomponent_skin == NULL) {
		g_warning("error creating builder's child");
		return FALSE;
	}
	gdam_builder_set_component(builder, subcomponent);
	if (skin->child_skin != NULL)
		gtk_object_unref(GTK_OBJECT(skin->child_skin));
	skin->child_skin = subcomponent_skin;
	gtk_object_ref(GTK_OBJECT(skin->child_skin));
	gtk_object_sink(GTK_OBJECT(skin->child_skin));
	if (skin->child_container) {
		gdam_debug("gdam_builder_skin: adding child to container");
		gdam_set_container(skin->child_container, 
				   subcomponent_skin->main_widget);
	}
	return TRUE;
}

static gboolean gdam_builder_skin_build_new_model(GdamSkin*    skin,
						      GdamChannel* channel,
						      int          num_options,
	                                              char**       options)
{
	GdamModel* model;
	model = GDAM_MODEL(gtk_type_new(GDAM_TYPE_BUILDER));
	model->channel = channel;
	skin->model = model;
	if (!options || num_options == 0) {
		return TRUE;
	}
	if (num_options < 2) {
		g_warning("making builder: too few arguments "
				"(need skin_type:filename)");
		return FALSE;
	}
	return gdam_builder_skin_set_child(
					GDAM_BUILDER_SKIN(skin),
					options[0],
					options[1],
					num_options - 2,
					options + 2);
}


static GdamSkinGenericConnect connect_infos[] =
{
	{ "gdam_builder_set_child",
	   gdam_builder_set_child_handler	},
	{ "gdam_builder_launch_filter",
	   gdam_builder_launch_filter_handler	},
	{ "gdam_builder_filter_box",
	   gdam_builder_filter_box_handler	},		};

static void gdam_builder_skin_destroy(GtkObject* object)
{
	GdamBuilderSkin* builder_skin = GDAM_BUILDER_SKIN(object);
	g_hash_table_destroy(builder_skin->filter_boxes);
	if (builder_skin->child_skin != NULL) {
		GdamSkin* child = builder_skin->child_skin;
		builder_skin->child_skin = NULL;
		gtk_object_unref(GTK_OBJECT(child));
	}
	(*parent_class->destroy)(object);
}

static void gdam_builder_skin_class_init(GdamBuilderSkinClass* klass)
{
	GdamSkinClass* skin_class = GDAM_SKIN_CLASS(klass);
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(klass);
	GdamViewClass* view_class = GDAM_VIEW_CLASS(klass);
	skin_class->build_new_model = gdam_builder_skin_build_new_model;
	object_class->destroy = gdam_builder_skin_destroy;
	view_class->synchronize = gdam_builder_skin_synchronize;
	gdam_skin_class_register_generic_handlers(skin_class,
			                          G_N_ELEMENTS(connect_infos),
			                          connect_infos);
}

gboolean gdam_builder_skin_move_up  (GdamBuilderSkin* 	skin,
				     GdamSkin*          filter_skin)
{
	GdamBuilder* builder = GDAM_BUILDER(GDAM_SKIN(skin)->model);
	GdamFilter* filter = GDAM_FILTER(filter_skin->model);
	GtkWidget* filter_top;
	int old_index = g_list_index(builder->filters_in, filter);
	int new_index;
	if (old_index <= 0)
		return TRUE;
	new_index = old_index - 1;
	gtk_widget_hide(filter_skin->main_widget);
	filter_top = g_hash_table_lookup(skin->filter_boxes, filter_skin);
	gtk_box_reorder_child(skin->filterbox,
	                      filter_top,
			      new_index);
	gtk_widget_show(filter_skin->main_widget);
	gdam_builder_move_filter(builder, filter, new_index);
	return TRUE;
}

gboolean gdam_builder_skin_move_down(GdamBuilderSkin* 	skin,
				     GdamSkin*          filter_skin)
{
	GdamBuilder* builder = GDAM_BUILDER(GDAM_SKIN(skin)->model);
	GdamFilter* filter = GDAM_FILTER(filter_skin->model);
	GtkWidget* filter_top;
	int old_index = g_list_index(builder->filters_in, filter);
	int new_index;
	int num_filters = g_list_length(builder->filters_in);
	if (old_index >= num_filters - 1)
		return TRUE;
	new_index = old_index + 1;
	filter_top = g_hash_table_lookup(skin->filter_boxes, filter_skin);
	gtk_box_reorder_child(skin->filterbox,
	                      filter_top,
			      new_index);
	gdam_builder_move_filter(builder, filter, new_index);
	return TRUE;
}

static GdamSkin* find_skin(GdamModel* filter)
{
	GSList* list;
	for (list = filter->skin_list; list != NULL; list = list->next)
		if (GDAM_IS_SKIN(list->data))
			return GDAM_SKIN(list->data);
	return NULL;
}
static gboolean prune_nonexistant(gpointer subskin,
                                  gpointer filterbox,
                                  gpointer valid_hashtable)
{
	GHashTable* valid = valid_hashtable;
	if (g_hash_table_lookup(valid, subskin) != NULL)
		return FALSE;
	return TRUE;
}


static void gdam_builder_skin_synchronize(GdamView* comp_skin)
{
	GdamBuilderSkin* builder_skin = GDAM_BUILDER_SKIN(comp_skin);
	GdamSkin* skin = GDAM_SKIN(comp_skin);
	GdamBuilder* builder = GDAM_BUILDER(skin->model);
	GList* list;
	GHashTable* new_filter_skins = g_hash_table_new(NULL, NULL);

	GSList* new_widget_list = NULL;

	for (list = builder->filters_in; list != NULL; list = list->next) {
		GdamModel* filter = list->data;
		GdamSkin* sub_skin = find_skin(filter);
		GtkWidget* widget;
		if (sub_skin == NULL) {
			g_warning("couldn't find skin for component [%s]",
			      gtk_type_name(GTK_OBJECT(filter)->klass->type));
			continue;
		}
		g_hash_table_insert(new_filter_skins, sub_skin, sub_skin);
		widget = g_hash_table_lookup(builder_skin->filter_boxes,
		                             sub_skin);
		if (widget == NULL) {
			widget = make_widget_for_filter(builder_skin,
			                                sub_skin,
							NULL);
		}
		new_widget_list = g_slist_prepend(new_widget_list, widget);
		gtk_object_ref((GtkObject*)widget);
	}
	new_widget_list = g_slist_reverse(new_widget_list);
	gdam_gtk_container_remove_all(GTK_CONTAINER(builder_skin->filterbox));
	while (new_widget_list != NULL) {
		GtkWidget* widget = new_widget_list->data;
		new_widget_list = g_slist_remove(new_widget_list, widget);
		gtk_box_pack_start_defaults(builder_skin->filterbox, widget);
		gtk_object_unref(GTK_OBJECT(widget));
	}
	g_hash_table_foreach_remove(builder_skin->filter_boxes,
	                            prune_nonexistant, 
			            new_filter_skins);
	g_hash_table_destroy(new_filter_skins);

	if (builder->component != NULL) {
		GdamSkin* subcomponent_skin;
		GdamModel* child = builder->component;
		subcomponent_skin = gdam_skin_find_from_model(child);
		if (subcomponent_skin != NULL) {
			if (builder_skin->child_skin != NULL)
				gtk_object_unref(GTK_OBJECT(
						builder_skin->child_skin
							));
			builder_skin->child_skin = subcomponent_skin;
			gtk_object_ref(GTK_OBJECT(builder_skin->child_skin));
			gtk_object_sink(GTK_OBJECT(builder_skin->child_skin));
			if (builder_skin->child_container) {
				gdam_set_container(builder_skin->
							child_container, 
						   subcomponent_skin->
						   	main_widget);
			}
		}
	}
}

